## -*- texinfo -*-
## @deftypefn {Function File} {@var{p} =} parse (@var{p}, @var{varargin})
## Parse the arguments as specified by the inputParser object, @var{p}.
## @end deftypefn

## Author: Ben Abbott <bpabbott@mac.com>
## Created: 2010-11-15

function p = parse (p, varargin)
  plan = get (p, "plan");
  results = get (p, "Results");
  using_defaults = {};
  if (get (p, "CaseSensitive"))
    name_cmp = @strcmp;
  else
    name_cmp = @strcmpi;
  endif
  if (get (p, "StructExpand"))
    k = find (cellfun (@isstruct, varargin));
    for m = numel(k):-1:1
      n = k(m);
      s = varargin{n};
      c = [fieldnames(s).'; struct2cell(s).'](:).';
      if (n > 1 && n < numel (varargin))
        varargin = horzcat (varargin(1:n-1), c, varargin(n+1:end));
      elseif (numel (varargin) == 1)
        varargin = c;
      elseif (n == 1);
        varargin = horzcat (c, varargin(n+1:end));
      else # n == numel (varargin)
        varargin = horzcat (varargin(1:n-1), c);
      endif
    endfor
  endif
  if (isempty (results))
    results = struct ();
  endif
  for n = 1 : numel (plan)
    found = false;
    results.(plan(n).name) = plan(n).default;
    if (! isempty (varargin))
      switch plan(n).type
      case "required"
        found = true;
        value = varargin{1};
        varargin(1) = [];
      case "optional"
        m = find (cellfun (@ischar, varargin));
        k = find (name_cmp (plan(n).name, varargin(m)));
        if (isempty (k) && validate_arg (plan(n).validator, varargin{1}))
          found = true;
          value = varargin{1};
          varargin(1) = [];
        elseif (! isempty (k))
          m = m(k);
          found = true;
          value = varargin{max(m)+1};
          varargin(union(m,m+1)) = [];
        endif
      case "paramvalue"
        m = cellfun (@ischar, varargin);
        k = find (name_cmp (plan(n).name, varargin(m)));
        if (! isempty (k))
          found = true;
          m = m(k);
          value = varargin{max(m)+1};
          varargin(union(m,m+1)) = [];
        endif
      otherwise
        error ("inputParser:parse",
               "parse (inputParser): Invalid argument type.")
      endswitch
    endif
    if (found)
      if (validate_arg (plan(n).validator, value))
        results.(plan(n).name) = value;
      else
        error ("inputParser:invalidinput",
               "inputPareser: input '%s' is invalid.\n", plan(n).name);
      endif
      p = set (p, "Parameters", union (get (p, "Parameters"), {plan(n).name}));
    elseif (strcmp (plan(n).type, "required"))
      error ("inputParser:missinginput",
             "inputPareser: input '%s' is missing.\n", plan(n).name);
    else
      using_defaults = union (using_defaults, {plan(n).name});
    endif
  endfor

  if (! isempty (varargin))
    if (get (p, "KeepUnmatched"))
    ## Include properties that do not match specified properties
      for n = 1:2:numel(varargin)
        if (ischar (varargin{n}))
          results.(varargin{n}) = varargin{n+1};
        else
          error ("inputParser:invalidinput", "inputPareser: invalid input")
        endif
      endfor
    endif
    p = set (p, "Unmatched", varargin{1:2:end});
  endif

  ## Store the results of the parsing
  p = set (p, "Results", results);
  p = set (p, "UsingDefaults", using_defaults);

  ## FIXME - Identify arguments using defaults (UsingDefaults)
endfunction

function result = validate_arg (validator, arg)
  try
    result = validator (arg);
  catch
    result = false;
  end_try_catch
endfunction

